# Copyright 2011 Tommy Poulter
#
# This file is part of basechip.
#
# basechip is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
# 
# basechip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with basechip.  If not, see <http://www.gnu.org/licenses/>.

require 'base_chip/menu'
module BaseChip
  class TopMenu < Menu
    include Cli
    def ready_workload(*args)
      @targets = args
      BaseChip.load_environment
      @project = BaseChip.project
      @project.ready(@targets) if @targets.size > 0
    end

    option :yes  , :description=>"Accept defaults on any posed questions", :type=>:boolean, :alias=>[:y]
    option :modes, :description=>"Enable predefined modes"               , :type=>:string , :alias=>[:m]

    desc "Create a new project"
    option :git, :description=>"Set up as a git repository", :type=>:boolean, :alias=>[:g,:r,:repo]
    def init(project_name, shared_directory)
      $shared_directory = shared_directory
      BaseChip.project  = project_name
      BaseChip.root     = "./#{project_name}"
      FileUtils.mkdir_p BaseChip.root
      BaseChip.build_directory_structure(BaseChip.root, BaseChip.project_directories)
      BaseChip.erb_template "#{BaseChip::GEM_DIR}/collateral/project/settings.rb.erb", "#{BaseChip.root}/base_chip/settings.rb"
      BaseChip.erb_template "#{BaseChip::GEM_DIR}/collateral/project/project.rb.erb" , "#{BaseChip.root}/base_chip/project.rb"
      FileUtils.cp          "#{BaseChip::GEM_DIR}/collateral/project/gitignore"      , "#{BaseChip.root}/.gitignore"
      FileUtils.cd          BaseChip.root

      if options.git
        system "git init ."
        system "git add ./base_chip ./.gitignore"
        system "git commit -m 'Initial automated project build by BaseChip.' ."

        normal "Your project has been created.\n" +
               "Your next step is probably to generate a block.\n" +
               "You can learn more by entering your newly created directory, and typing:\n" +
               "  basechip help generate block"
      end
    end

    desc "Generate a block, tool, testbench, etc"
    sub_command :generate, GeneratorMenu

    desc  'Install a recipe, plugin, or tool'
    sub_command :install, InstallMenu

    desc "List blocks, actions, tools, tests, etc"
    long "If 'pattern' is given as a string, it must completely match the name of the block/action/test/test_list\n" +
         "    or one of it's parents.  If surrounded by '/', the pattern will be considered a regular expression\n"  +
         "    and will be compared to the full name of the item including it's parents' names separated by colons."
    sub_command :list, ListMenu

    desc  'Execute targets'
    full_usage 'basechip <targets+> [options]'
    default
    option :work_dir       , :description=>"Working directory other than cwd"     , :type=>:string , :alias=>[:w]
    option :max_fails      , :description=>"Working directory other than cwd"     , :type=>:string , :alias=>[:f]
    option :max_passes     , :description=>"Working directory other than cwd"     , :type=>:string , :alias=>[:p]
    option :repeat         , :description=>"Repeat each test 'r' times"           , :type=>:integer, :alias=>[:r]
    option :jobs           , :description=>"Number of jobs, and for which cluster", :type=>:string , :alias=>[:j]
    option :seed           , :description=>"Root seed for all actions"            , :type=>:integer, :alias=>[:s]
    option :client_of      , :description=>"Internal use only"                    , :type=>:string
    option :client_id      , :description=>"Internal use only"                    , :type=>:integer 
    option :opc            , :description=>"Internal use only"                    , :type=>:string
    option :opi            , :description=>"Internal use only"                    , :type=>:string
    def exec(*args)
      begin
        fault "Repeat is a planned feature" if BaseChip.options.repeat

        args << 'all' if options.client_of && args.size == 0
        ready_workload(*args)

        @project.enable_tracking(options.opc,options.opi) if options.opc
        @project.tasker.clusters = @project.dereference_clusters
        @project.tasker.client_of = options.client_of
        @project.tasker.client_id = options.client_id || 0

        if options.client_of
          @project.tasker.run
        else
          if @targets.size > 0
            begin
              @project.run_all_readied
              # @project.track_complete 'pass'
            rescue Reporting::Fault => e
              # @project.track_complete 'fail', e.message
              exit 1
            rescue RuntimeError => e
              # @project.track_complete 'fail', e.message
              p e
              raise e
            end
          end
        end
      rescue Interrupt # quetly handle the interrupt of clients
        raise unless options.client_of 
      end
    end

    desc  'rerun targets.  Some output remains'
    def rerun(bundle_path)
      fault "#{bundle_path} does not exist" unless File.exist?(bundle_path)
      case bundle_path
      when /(.*)\.tgz$/ 
        puts "Unzipping #{bundle_path}"
        system "tar -zxvf #{bundle_path}"
        bundle_path = $1
      end
      fault "#{bundle_path}/rerun.yaml does not exist" unless File.exist?("#{bundle_path}/rerun.yaml")
      rerun_hash = YAML.load(File.read("#{bundle_path}/rerun.yaml"))
      BaseChip.options.seed            ||= rerun_hash['seed' ]
      BaseChip.options.modes           ||= rerun_hash['modes']
      BaseChip.options. append_arguments = (BaseChip.options. append_arguments || []) + (rerun_hash[' append_arguments'] || [])
      BaseChip.options.replace_arguments = (BaseChip.options.replace_arguments || []) + (rerun_hash['replace_arguments'] || [])
      BaseChip.options. append_arguments = nil if BaseChip.options. append_arguments.empty?
      BaseChip.options.replace_arguments = nil if BaseChip.options.replace_arguments.empty?
      ready_workload(rerun_hash['full_name'])
      @project.workload.first.directory "#{@project.workload.first.directory}.rerun"
      @project.run_all_readied
    end

    desc  'Clean targets.  Some output remains'
    usage '<targets+> [options]'
    def clean(*args)
      args << 'all' if args.empty?
      ready_workload(*args)
      @project.workload.each do |w|
        w.clean
      end
    end

    desc  'Destroy all output relate to targets'
    usage '<targets+> [options]'
    def clobber(*args)
      args << 'all' if args.empty?
      ready_workload(*args)
      @project.workload.each do |w|
        w.clobber
      end
    end

    # desc  'Return completion options for a command'
    # def complete
    #   args    = BaseChip.options. append_arguments
    #   current = BaseChip.options.replace_arguments[0]
    #   args.shift

    #   out = []
    #   tasks = TopMenu.cli.tasks.keys.map{|s|s.to_s}

    #   # if args.empty?
    #   #   if current
    #   #     tasks = TopMenu.cli.tasks.keys.map{|s|s.to_s}
    #   #     if tasks.include? current
    #   #       puts current
    #   #       return
    #   #     end
    #   #   else
    #   #     if tasks.include? current
    #   #       puts current
    #   #       return
    #   #     end
    #   #   end
    #   # end
    #   # case args[0]
    #   # when (*%w{clean clobber})
    #   #   out += BaseChip.blocks.map{|b|b.name}
    #   # else
    #   #   ready_workload('all')
    #   #   out += TopMenu.cli.tasks.keys
    #   #   out += BaseChip.blocks.map{|b|b.name}
    #   # end

    #   if self.class.workload_commands.include? argv[0]
    #     ready_workload('all')
    #     if current
    #     else
    #       out += BaseChip.blocks.map{|b|b.name.to_s}.keep_if{|b| b=~ /^#{current}/}
    #     end
    #   end
    #   puts out.join(" ")
    # end
    # def self.workload_commands
    #   return @@workload_commands if @@workload_commands
    #   @@workload_commands = [nil]
    #   self.cli.tasks.each do |k,v|
    #     @@workload_commands << k.to_s if v.usage =~ /targets/
    #   end
    #   @@workload_commands
    # end
  end
end
